/* vim: set foldmethod=marker : */

#include <stdint.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/boot.h>
#include "hm2007.h"
#include "commands.h"
#include "uart.h"
#include "cc2500.h"



static prog_char STR_TALK[] = "+ Talk\n";
static prog_char STR_DONE[] = "+ Done\n";
static prog_char STR_HM2007[] = "# HM2007\n";
static prog_char STR_END[] = "# END\n";
static prog_char STR_SEND_FILE[] = "+ Send file\n";
static prog_char STR_FORMAT_ERROR[] = "\n! File format error\n";
static prog_char STR_AUTO_ON[] = "\n+ Switched to auto mode. Press any key to return to command mode\n";
static prog_char STR_AUTO_OFF[] = "\n+ Returned to command mode\n\n";

static prog_char STR_HELP[] = "\
      ==== Commands ====\n\
d Download patterns to file\n\
u Upload patterns from file\n\
t [number] Train pattern [number]\n\
r Recognize pattern\n\
l [0/1] Set pattern length (0 = 0.96s, 1= 1.9s)\n\
h or ? This help message\n\
";

static prog_char STR_BANNER[] = "\
\n\n****************************\n\
* HM2007 voice recognition *\n\
****************************\n\n";

static prog_char STR_MEMORY_OK[] = "Memory test OK\n";
static prog_char STR_PATTERNS_DONE[] = "Patterns uploaded\nSend ? for commands list\n\n";

static prog_char STR_NOT_RECOG[] = "! Command not recognized\n";
static prog_char STR_PARAMS[] = "! Invalid parameters\n";
static prog_char STR_CLEARING[] = "+ Clearing patterns, please wait ... ";
static prog_char STR_DONE2[] = "Done\n";
static prog_char STR_RECOG[] = "+ Recognition mode\n";
static prog_char STR_NO_MATCH[] = "! No match\n";



uint8_t word_len EEMEM = 1;
unsigned char patterns[256 * 40] PROGMEM = {};

static int uart_putchar(char c, FILE *stream);

static FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL,
										 _FDEV_SETUP_WRITE);


void write_pattern(unsigned char pattern, unsigned char *buf) BOOTLOADER_SECTION;

void write_pattern(unsigned char pattern, unsigned char *buf) /*{{{*/
{
	unsigned char sreg;
	uint32_t addr, page_addr, page_end;
	uint16_t len = buf[0] + (buf[1] << 8);

	
	addr = (uint32_t)(uint16_t)(patterns + pattern * 256);
	page_addr = (addr & ~(SPM_PAGESIZE - 1));
	page_end = ((addr + len + 2) & ~(SPM_PAGESIZE - 1)) + SPM_PAGESIZE;

	sreg = SREG;
	cli();

	eeprom_busy_wait();

	while (page_addr < page_end)
	{
		if (page_addr < addr || page_addr >= addr + len + 2) {
			uint16_t w = pgm_read_word(page_addr);
			boot_page_fill(page_addr, w);
		} else {
			uint16_t w = ((uint16_t*)buf)[(page_addr - addr) >> 1];
			boot_page_fill(page_addr, w);
		}
		if ((page_addr & (SPM_PAGESIZE - 1)) == (SPM_PAGESIZE - 2)) {
			boot_page_erase(page_addr);
			boot_spm_busy_wait();
			boot_page_write (page_addr);
			boot_spm_busy_wait();
		}
		page_addr += 2;
	}
	
	boot_rww_enable ();
	if (sreg & 0x80) {
		sei();
	}
}/*}}}*/

static int uart_putchar(char c, FILE *stream)/*{{{*/
{
	if (c == '\n') {
		uart_putchar('\r', stream);
	}
	uart_tx(c);
	return 0;
}/*}}}*/

void printp(prog_char *str)
{
	char c;
	while ((c = pgm_read_byte(str))) {
		uart_putchar(c, 0);
		str++;
	}
}

void train(int pattern)/*{{{*/
{
	uint8_t max;
	if (eeprom_read_byte(&word_len)) {
		max = 20;
	} else {
		max = 40;
	}
	if (pattern < 1 || pattern > max) {
		printf("! Number out of range. Specify a number from 1 to %d\n", max);
		return;
	}
	printf("+ Training pattern #%d\n", pattern);
	hm_train(pattern);
	printp(STR_TALK);
	while  (!hm_ready());

	unsigned int len = hm_pattern_size(pattern);
	unsigned char buf[300];
	hm_download(len, buf + 2);
	buf[0] = len & 0xFF;
	buf[1] = len >> 8;

	uart_xoff();
	_delay_ms(10);
	_delay_ms(10);
	write_pattern(pattern-1, buf);
	uart_xon();

	printp(STR_DONE);
}/*}}}*/

void download(void)/*{{{*/
{
	int i;
	uint8_t max;
	uint16_t addr;
	if (eeprom_read_byte(&word_len)) {
		max = 20;
	} else {
		max = 40;
	}
	printp(STR_HM2007);
	printf("%02X\n", eeprom_read_byte(&word_len));
	for (i = 0; i < max; i++) {
		int j;
		uint16_t len;
		addr = (uint16_t)(patterns + i * 256);
		len = pgm_read_word(addr);
		printf("%02X", len);
		for (j = 0; j < len; j++) {
			if (!(j & 0x1F)) {
				uart_putchar('\n', 0);
			}
			printf("%02X", pgm_read_byte(addr + j + 2));
		}
		uart_putchar('\n', 0);
	}
	printp(STR_END);
}/*}}}*/

static unsigned int hexint(unsigned char *hex)/*{{{*/
{
	if (
		((hex[0] >= '0' && hex[0] <= '9') || (hex[0] >= 'A' && hex[0] <= 'F'))
		||
		((hex[1] >= '0' && hex[1] <= '9') || (hex[1] >= 'A' && hex[1] <= 'F'))
	   ) {
		unsigned int ret;
		if (hex[0] < 'A') {
			ret = hex[0] - '0';
		} else {
			ret = hex[0] - 'A' + 0x0A;
		}
		ret <<= 4;
		if (hex[1] < 'A') {
			ret |= (hex[1] - '0');
		} else {
			ret |= (hex[1] - 'A' + 0x0A);
		}
		return ret;
	}
	return 0x100;
}/*}}}*/

static void rx2(unsigned char *buf)/*{{{*/
{
	uint8_t i = 0;
	char c;
	for (;;) {
		c = uart_rx();
		if (c == '\r' || c == '\n') continue;
		buf[i] = c;
		i++;
		if (i == 2) break;
	}
}/*}}}*/

void upload(void)/*{{{*/
{
	printp(STR_SEND_FILE);
	unsigned char buf[2], c;
	int len, max, j;

	do {
		c = uart_rx();
	} while (c != '#');

	for (;;) {
		c = uart_rx();
		if (c == '\r') continue;
		if (c == '\n') break;
	}

	rx2(buf);
	len = hexint(buf);
	if (len == 0x100) {
		printp(STR_FORMAT_ERROR);
		return;
	}
	eeprom_busy_wait();
	eeprom_write_byte(&word_len, len != 0);
	if (len) {
		max = 20;
	} else {
		max = 40;
	}
	for (j=0; j < max; j++) {
		unsigned char data[256];
		int i;
		rx2(buf);
		len = hexint(buf);
		if (len == 0x100) {
			printp(STR_FORMAT_ERROR);
			return;
		}
		for(i = 0; i < len; i++) {
			uint16_t byte;
			rx2(buf);
			byte = hexint(buf);
			if (byte == 0x100) {
				printp(STR_FORMAT_ERROR);
				return;
			}
			data[2 + i] = byte;
		}
		data[0] = len & 0xFF;
		data[1] = len >> 8;
		uart_xoff();
		_delay_ms(10);
		_delay_ms(10);
		write_pattern(j, data);
		uart_xon();
	}
	while (uart_rx() != '\n');
	while (uart_rx() != '\n');
	printp(STR_DONE);
	while (rx_avail()) {
		uart_rx();
	}
}/*}}}*/

void help(void)
{
	printp(STR_HELP);
}

int main(void)/*{{{*/
{

	/*  SPI */
	DDRB = 0xB0;
	PORTB |= _BV(6) | _BV(4);
	SPCR = _BV(SPE) | _BV(MSTR);

	_delay_ms(10);
	_delay_ms(10);

	/*     */
	cc2500_por();
	/*   ,  SmartRF Studio */
	cc2500_init();

	/*     */
	cc2500_write(CC2500_PATABLE, 0xFF);

	cc2500_write(CC2500_MCSM0,	0x18); // Calibration from IDLE to TX/RX
	cc2500_write(CC2500_MCSM1,	0x00); // No CCA, IDLE after TX and RX
	cc2500_write(CC2500_PKTCTRL0, 0x45); // Enable data whitening and CRC
	cc2500_write(CC2500_PKTCTRL1, 0x04); // Enable append mode

	init_uart();
	stdout = &mystdout;

	sei();

	printp(STR_BANNER);

	hm_init();

	_delay_ms(10);
	_delay_ms(10);
	_delay_ms(10);
	hm_wlen(eeprom_read_byte(&word_len));
	hm_reset();
	printp(STR_MEMORY_OK);

	uint16_t i;
	for (i = 0; i < 20; i ++) {
		uint16_t len = pgm_read_word(patterns + i *256);
		if (len) {
			int j;
			hm_start_upload(i+1, len);
			for (j = 0; j < len; j++) {
				hm_write_byte(pgm_read_byte(patterns + i*256+j+2));
			}
		}
	}

	printp(STR_PATTERNS_DONE);

	printp(STR_AUTO_ON);

	char cmd = CMD_WAIT;
	uint8_t txbuf[4];
	uint8_t pcktnum = 1;
	for (;;) {
		unsigned int param;
		if (cmd == CMD_WAIT) {
			if (rx_avail()) {
				while (rx_avail()) {
					uart_rx();
				}
				cmd = CMD_AUTO;
				printp(STR_AUTO_OFF);
			} else {
				hm_recog();
				while (!hm_ready());
				unsigned int result = hm_result();

				/*     -   */
				txbuf[0] = 3;

				/*   */
				txbuf[1] = pcktnum;
				txbuf[2] = result & 0xFF;
				txbuf[3] = result >> 8;

				pcktnum++;
				if (result >= 0xFF00) {
					printp(STR_NO_MATCH);
				} else {
					uint8_t i;
					/*  10     */
					for (i = 0; i < 10; i++) {
						/*    */
						cc2500_write_burst(CC2500_TX_FIFO, 4,  txbuf);
						/*   */
						cc2500_command(CC2500_STX);
						/*    */
						while ((cc2500_status(0) & 0xF0) != 0) {
							;
						}
						_delay_ms(10);
					}
					printf("+ Result : %d. Distance: %d\n", result & 0xFF, result >> 8);
				}
			}
			continue;
		}
		get_command(&cmd, &param);
		switch (cmd) {
			case ERROR_UNKNOWN_CMD :
				printp(STR_NOT_RECOG);
				break;
			case ERROR_PARAMS:
				printp(STR_PARAMS);
				break;
			case CMD_TRAIN:
				train(param);
				break;
			case CMD_AUTO:
				printp(STR_AUTO_ON);
				cmd = CMD_WAIT;
				break;
			case CMD_DOWNLOAD:
				download();
				break;
			case CMD_UPLOAD:
				upload();
				break;
			case CMD_LENGTH:
				if ((param ? 1 :0) != (eeprom_read_byte(&word_len) ? 1 : 0)) {
					if (!confirm("All patterns will be cleared. Continue?")) {
						break;
					}
					printp(STR_CLEARING);
					while (!tx_empty());
					uart_xoff();
					_delay_ms(10);
					_delay_ms(10);
					hm_wlen(param ? 1 : 0);
					eeprom_write_byte(&word_len, param ? 1 : 0);
					int i;
					uint16_t buf[1];
					for (i = 0; i < 40 ; i++) {
						*buf = 0;
						write_pattern(i, (unsigned char *)buf);
					}
					hm_reset();
					uart_xon();
					printp(STR_DONE2);
				}
				break;
			case CMD_RECOG:
				printp(STR_RECOG);
				printp(STR_TALK);

				hm_recog();
				while (!hm_ready());
				unsigned int result = hm_result();
				if (result >= 0xFF00) {
					printp(STR_NO_MATCH);
				} else {
					printf("+ Result : %d. Distance: %d\n", result & 0xFF, result >> 8);
				}
				break;
			case CMD_HELP:
				help();
				break;
		}
	}
}/*}}}*/

