#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <stdio.h>
#include "uart.h"

#define XON 0x11
#define XOFF 0x13

#define PROMOTE(x) {x++; if (x >= UART_BUF_SIZE) x = 0; }
#define TX_ON() {UCSRB |= _BV(UDRIE); }
#define TX_OFF() {UCSRB &= ~_BV(UDRIE); }
#define TX_URGENT(c) {tx_urgent = c; TX_ON(); }

#define XOFF_BUFFER 1
#define XOFF_FORCE 2

static volatile char rxbuf[UART_BUF_SIZE];
static volatile uint8_t rxhead = 0, rxtail = 0;
static volatile uint16_t rxread = 0, rxwrite = 0;

static volatile char txbuf[UART_BUF_SIZE];
static volatile uint8_t txhead = 0, txtail = 0;
static volatile uint16_t txread = 0, txwrite = 0;

volatile char tx_urgent = 0;
volatile char cantx = 1, canrx = 0;

void init_uart(void)/*{{{*/
{
	UCSRB = _BV(TXEN) |_BV(RXEN) | _BV(UDRIE) | _BV(RXCIE);
	UCSRC = _BV(UMSEL) | (2 << UCSZ0);
	UBRRH = 0;
	UBRRL = 25;
}/*}}}*/


ISR(SIG_USART_RECV)
{
	char c = UDR;
	if (c == XON) {
		TX_ON();
		cantx = 1;
	} else 	if (c == XOFF) {
		TX_OFF();
		cantx = 0;
	} else {
		if (rxwrite - rxread < UART_BUF_SIZE) {
			rxbuf[rxhead] = c;
			PROMOTE(rxhead);
			rxwrite++;
			if ((rxwrite - rxread) > (UART_BUF_SIZE / 2)) {
				TX_URGENT(XOFF);
				if (!canrx) {
					canrx = XOFF_BUFFER;
				}
			}
		}
	}
}

ISR(SIG_USART_DATA)
{
	if (tx_urgent) {
		UDR = tx_urgent;
		tx_urgent = 0;
		return;
	}
	if (txwrite - txread) {
		UDR = txbuf[txtail];
		PROMOTE(txtail);
		txread++;
	} else {
		TX_OFF();
	}
}

uint8_t rx_avail(void)
{
	return rxwrite - rxread;
}

uint16_t tx_avail(void)
{
	return UART_BUF_SIZE - (txwrite - txread);
}

char uart_rx(void)
{
	/*
	while (!(UCSRA & _BV(RXC)));
	return UDR;
	*/
	char ret;
	while (!(rxwrite - rxread));
	ret = rxbuf[rxtail];
	PROMOTE(rxtail);
	rxread++;
	if ((rxwrite - rxread) < (UART_BUF_SIZE / 4)) {
		if (canrx == XOFF_BUFFER) {
			TX_URGENT(XON);
			canrx = 0;
		}
	}
	return ret;
}

void uart_tx(char c)
{
	/*
	while (!(UCSRA & _BV(UDRE)));
	UDR = c;
	return;
	*/
	while (txwrite - txread >= UART_BUF_SIZE);
	txbuf[txhead] = c;
	PROMOTE(txhead);
	txwrite++;
	if (cantx) {
		TX_ON();
	}
}

void uart_xoff(void)
{
	while (tx_urgent);
	TX_URGENT(XOFF);
	canrx = XOFF_FORCE;
}

void uart_xon(void)
{
	TX_URGENT(XON);
	canrx = 0;
}

char tx_empty(void)
{
	return rxread == rxwrite;
}

